//
// decimal_io.hpp
//
// Bill Seymour, 2024-02-02
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This header provides the definitions of the << and >> I/O operators
// for the decimal class defined in decimal.hpp, in which file this header
// has already been #included if necessary.
//

#ifndef DECIMAL_IO_HPP_INCLUDED
#define DECIMAL_IO_HPP_INCLUDED

#include "decimal_parse.hpp"

#include <iostream>
#include <locale>

namespace bignum {

template<class Ch, class Tr>
std::basic_ostream<Ch,Tr>&
operator<<(std::basic_ostream<Ch,Tr>& ostr, const decimal& rhs)
{
    //
    // The << operator is documented to normalize the value, but
    // decimal::normalize() is non-const, so we'll actually output
    // a temporary object; and we might as well eagerly set val's
    // scale to the requested precision while we're at it.
    //
    decimal val(rhs);

    using iob = std::ios_base;

    const iob::fmtflags flags = ostr.flags();
    const bool sci = (flags & iob::scientific) != 0;

    using ctype_facet = std::ctype<Ch>;
    using punct_facet = std::numpunct<Ch>;

    const std::locale& loc = ostr.getloc();
    const ctype_facet& cf = std::use_facet<ctype_facet>(loc);
    const punct_facet& pf = std::use_facet<punct_facet>(loc);

    char sign = val.isneg() ? '-' : (flags & iob::showpos) ? '+' : '\0';
    std::basic_string<Ch,Tr> sval;

    if (val.iszero())
    {
        sval.push_back(cf.widen('0'));
    }
    else
    {
        std::string temp(val.abs().to_string(int(ostr.precision()), sci));

        //
        // Get the string in reverse order.  This makes it easier
        // to insert thousands separators and to add the sign later.
        // We'll reverse the characters when we actually write the string.
        //
        bool got_point = false;
        //std::size_t outcnt = temp.size();
        using cit = std::string::reverse_iterator;
        for (cit it = temp.rbegin(), end = temp.rend(); it != end; ++it)
        {
            if ('.' == *it)
            {
                sval.push_back(pf.decimal_point());
                got_point = true;
            }
            else if (sci && (flags & iob::uppercase) && 'e' == *it)
            {
                sval.push_back(cf.widen('E'));
            }
            else
            {
                sval.push_back(cf.widen(*it));
            }
        }

        const std::string& grp = pf.grouping();
        std::string::const_iterator grp_it = grp.begin();
        const std::string::const_iterator grp_end = grp.end();

        char grp_len = CHAR_MAX; // number of characters allowed in a group
        if (grp_it != grp_end && *grp_it > 0)
            grp_len = *grp_it++;
        char grp_save = grp_len;

        char grp_cnt = 0; // number of characters so far in a group

        if (grp_save != CHAR_MAX)
        {
            //
            // Add thousands separators forward from the decimal point
            // (because we're in reverse order):
            //
            std::size_t point = sval.find(pf.decimal_point());
            std::size_t idx = std::string::npos == point ? 0 : point + 1;
            for ( ; idx < sval.size(); ++idx)
            {
                if (grp_len != CHAR_MAX && ++grp_cnt > grp_len)
                {
                    sval.insert(idx, 1, pf.thousands_sep());
                    grp_cnt = 0;

                    if (grp_it != grp_end && *grp_it > 0)
                        grp_len = *grp_it++;
                }
            }
        }
    }

    //
    // At this point, we have the desired string in reverse order in sval,
    // but no fill character(s) or sign.
    //

    int fill_width = int(ostr.width(0) - sval.size());
    Ch  fill_char  = ostr.fill();

    //
    // If the adjustfield flag bits don't specify exactly left or internal,
    // then we want to do right adjustment (left fill).
    //
    iob::fmtflags adj_bits = flags & iob::adjustfield;
    if (adj_bits != iob::left && adj_bits != iob::internal)
        adj_bits = iob::right;

    if (sign != '\0')
    {
        --fill_width;
        if (adj_bits == iob::right)
        {
            for ( ; fill_width > 0; --fill_width)
                ostr.put(fill_char);
        }
        ostr.put(cf.widen(sign));
    }

    //
    // At this point, we have actually written any required sign; and
    // if such was required, we have done any required right adjustment;
    // so right adjustment and internal fill now amount to the same thing.
    //
    if (adj_bits != iob::left)
    {
        for ( ; fill_width > 0; --fill_width)
            ostr.put(fill_char);
    }

    using iter = typename std::basic_string<Ch,Tr>::const_reverse_iterator;
    for (iter it = sval.crbegin(), end = sval.crend(); it != end; ++it)
        ostr.put(*it);

    for ( ; fill_width > 0; --fill_width)
        ostr.put(fill_char);

    if ((flags & iob::unitbuf) != 0)
        ostr.flush();

    return ostr;
}

template<class Ch, class Tr>
std::basic_istream<Ch,Tr>&
operator>>(std::basic_istream<Ch,Tr>& istr, decimal& value)
{
    using iob = std::ios_base;

    using ctype_facet = std::ctype<Ch>;
    using punct_facet = std::numpunct<Ch>;

    const std::locale& loc = istr.getloc();
    const ctype_facet& cf = std::use_facet<ctype_facet>(loc);
    const punct_facet& pf = std::use_facet<punct_facet>(loc);

    const bool dont_allow_grouping =
        pf.grouping().empty() || pf.grouping()[0] == CHAR_MAX;
    const Ch sep = pf.thousands_sep();
    const Ch pnt = pf.decimal_point();

    Ch wide_ch;
    if (!(istr >> wide_ch))
    {
        return istr;
    }

    std::string input; // narrowed validated input
    int state = 0;
    bool more = true;
    do {
        char c;
        if (sep == wide_ch)
        {
            if (!dont_allow_grouping)
                continue; // just ignore thousands separators
            c = '?';      // else parse should fail
        }
        else if (pnt == wide_ch)
        {
            c = '.'; // "C" locale decimal point
        }
        else
        {
            c = cf.narrow(wide_ch, '?');
        }

        state = parser::parse(state, c);

        switch (parser::action(state))
        {
            case parser::act::save:
                input.push_back(c);
                break;
/*
            case parser::act::done:
                input.push_back(c);
                more = false;
                break;
*/
            case parser::act::back:
                istr.putback(wide_ch);
                more = false;
                break;
            default: // parser::act::fail
                istr.setstate(iob::failbit);
                return istr;
        }
    } while(more && istr.get(wide_ch));

    if (more) // must be that istr.get() failed
    {
        if (istr.bad() || !istr.eof())
        {
            return istr;
        }

        // Just EOF...treat it like an unrecognized char:
        if (parser::action(parser::parse(state, '?')) != parser::act::back)
        {
            istr.setstate(iob::failbit);
            return istr;
        }
        // Else it's just the end of a valid value, so the stream is still good.
        istr.clear();
    }

    //
    // All we've done so far is to validate and narrow the input.
    // Use the decimal ctor to actually calculate the value.
    //
    value = decimal(input);
    return istr;
}

} // namespace bignum

#endif // DECIMAL_IO_HPP_INCLUDED
