//
// decimal_parse.cpp
//
// Bill Seymour, 2024-02-02
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This translation unit provides a parser to help the
// decimal class' >> istream operator validate its input.
// If all TUs in a program #define DECIMAL_NO_IO_OPERATORS,
// there's no need to link this code.
//

#include "decimal_parse.hpp"
#include <cctype> // isdigit

namespace {

//
// Token types:
//
#define DIG 0
#define SGN 1
#define PNT 2
#define E   3
#define OTH 4

inline int lex(char c)
{
    if (std::isdigit(static_cast<unsigned char>(c))) // most likely
    {
        return DIG;
    }
    switch(c)
    {
        case '-':
        case '+':
            return SGN;
        case '.':
            return PNT;
        case 'E':
        case 'e':
            return E;
        default:
            return OTH;
    }
}

//
// The parser is just a really simple finite state machine
// that doesn't do anything but validate the input.
//

// Ending states for the parser:
#define OK  8 // putback the char and we're done
#define ERR 9 // fail the stream

const int next_state[/*current state*/][/*token type*/OTH+1] =
{
  // state  DIG SGN PNT  E  OTH
    /*00*/ { 4 , 1 , 2 ,ERR,ERR},
    /*01*/ { 4 ,ERR, 2 ,ERR,ERR},
    /*02*/ { 3 ,ERR,ERR,ERR,ERR},
    /*03*/ { 5 ,ERR,ERR, 6 , OK},
    /*04*/ { 4 ,ERR, 2 , 6 , OK},
    /*05*/ { 5 ,ERR,ERR, 6 , OK},
    /*06*/ { 7 ,ERR,ERR,ERR,ERR},
    /*07*/ { 7 ,ERR,ERR,ERR, OK},
};

} // anonymous namespace

namespace bignum {
namespace parser {

int parse(int state, char c)
{
    return next_state[state][lex(c)];
}

// What to do in each state:
// enum class act { save, back, fail };
act action(int state)
{
    return state < OK ? act::save
                      : state < ERR ? act::back
                                    : act::fail;
}

} } // namespace bignum::parser

// End of decimal_parse.cpp
