//
// integer_io.hpp
//
// Bill Seymour, 2024-02-02
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This header provides the definitions of the << and >> I/O operators
// for the integer class defined in integer.hpp, in which file this header
// has already been #included if necessary.
//

#ifndef INTEGER_IO_HPP_INCLUDED
#define INTEGER_IO_HPP_INCLUDED

#include <iostream>
#include <locale>

namespace bignum {

template<typename Ch, typename Tr>
std::basic_ostream<Ch,Tr>&
operator<<(std::basic_ostream<Ch,Tr>& ostr, const integer& value)
{
    using iob = std::ios_base;

    const iob::fmtflags flags = ostr.flags();

    char sign = value.is_neg() ? '-' : (flags & iob::showpos ? '+' : '\0');

    int radix;
    switch (flags & iob::basefield)
    {
        case iob::oct:
            radix = 8;
            break;
        case iob::hex:
            radix = 16;
            break;
        default:
            radix = 10;
            break;
    }

    using ctype_facet = std::ctype<Ch>;
    using punct_facet = std::numpunct<Ch>;

    const std::locale& loc = ostr.getloc();
    const ctype_facet& cf = std::use_facet<ctype_facet>(loc);
    const punct_facet& pf = std::use_facet<punct_facet>(loc);

    const std::string& grp = pf.grouping();
    std::string::const_iterator grp_it = grp.cbegin();
    const std::string::const_iterator grp_end = grp.cend();

    char grp_len = CHAR_MAX; // number of characters allowed in a group
    if (grp_it != grp_end && *grp_it > 0)
        grp_len = *grp_it++;

    char grp_cnt = 0; // number of characters so far in a group

    std::string temp(value.to_string(radix));

    //
    // Widen and reverse the characters:
    //
    std::basic_string<Ch,Tr> sval;
    using cit = std::string::reverse_iterator;
    for (cit it = temp.rbegin(), end = temp.rend(); it != end; ++it)
    {
        sval.push_back(cf.widen(*it));
    }
    if (cf.widen('-') == sval.back())
    {
        sval.pop_back(); // We'll put the sign back later.
    }

    //
    // Note that we generated the characters in reverse order (LSD first)
    // and just appended them to the end of the string.  This makes it easy
    // to insert thousands separators and to add a sign or a base later.
    // We'll reverse the characters when we actually write the string.
    //
    for (std::size_t idx = 0; idx < sval.size(); ++idx)
    {
        if (grp_len != CHAR_MAX && ++grp_cnt > grp_len)
        {
            sval.insert(idx, 1, pf.thousands_sep());
            grp_cnt = 0;

            if (grp_it != grp_end && *grp_it > 0)
                grp_len = *grp_it++;
        }
    }

    //
    // At this point, we have the digit string with thousands separators
    // in reverse order in sval, but no fill character(s), sign or base.
    //

    int fill_width = int(ostr.width(0) - sval.size());
    Ch  fill_char  = ostr.fill();

    //
    // If the adjustfield flag bits don't specify exactly left or internal,
    // then we want to do right adjustment (left fill).
    //
    iob::fmtflags adj_bits = flags & iob::adjustfield;
    if (adj_bits != iob::left && adj_bits != iob::internal)
        adj_bits = iob::right;

    if (radix == 8)
    {
        if ((flags & iob::showbase) != 0 && !value.is_zero())
        {
            if (grp_cnt == grp_len)
            {
                --fill_width;
                sval.append(1, pf.thousands_sep());
            }
            --fill_width;
            sval.append(1, cf.widen('0'));
        }
    }
    else if (radix == 16)
    {
        //
        // integer::to_string() generates uppercase hex characters
        //
        if ((flags & iob::uppercase) == 0)
        {
            Ch* sbeg = const_cast<Ch*>(sval.c_str());
            cf.tolower(sbeg, sbeg + sval.size());
        }

        if ((flags & iob::showbase) != 0 && !value.is_zero() && sign == '\0')
        {
            fill_width -= 2;  // room for "0x"
            if (adj_bits == iob::right)
            {
                for ( ; fill_width > 0; --fill_width)
                    ostr.put(fill_char);
            }
            ostr.put(cf.widen('0'));
            ostr.put(cf.widen(flags & iob::uppercase ? 'X' : 'x'));
        }
    }

    if (sign != '\0')
    {
        --fill_width;
        if (adj_bits == iob::right)
        {
            for ( ; fill_width > 0; --fill_width)
                ostr.put(fill_char);
        }
        ostr.put(cf.widen(sign));
    }

    //
    // At this point, we have actually written any required "0x" or sign;
    // and if such was required, we have done any required right adjustment;
    // so right adjustment and internal fill now amount to the same thing.
    //
    if (adj_bits != iob::left)
    {
        for ( ; fill_width > 0; --fill_width)
            ostr.put(fill_char);
    }

    using iter = typename std::basic_string<Ch,Tr>::const_reverse_iterator;
    for (iter it = sval.crbegin(), end = sval.crend(); it != end; ++it)
        ostr.put(*it);

    for ( ; fill_width > 0; --fill_width)
        ostr.put(fill_char);

    if ((flags & iob::unitbuf) != 0)
        ostr.flush();

    return ostr;
}

template<typename Ch, typename Tr>
std::basic_istream<Ch,Tr>&
operator>>(std::basic_istream<Ch,Tr>& istr, integer& value)
{
    using iob = std::ios_base;

    const std::locale& loc = istr.getloc();

    using ctype_facet = std::ctype<Ch>;
    const ctype_facet& cf = std::use_facet<ctype_facet>(loc);

    Ch c;

    // First character, or first non-white character if skipws:
    if (!(istr >> c))
        return istr;

    bool isneg = false;
    if (cf.narrow(c, '\0') == '-')
    {
        isneg = true;
        if (!istr.get(c))
            return istr;
    }
    else if (cf.narrow(c, '\0') == '+')
    {
        if (!istr.get(c))
            return istr;
    }

    //
    // There must be at least one valid digit for integer input, otherwise
    // the stream is set to failed.  Also, as is the case for operator<< above,
    // if the base isn't exactly oct or hex, then it's dec; and there's no way
    // for the user to type a "0" or "0x" prefix to select the base.
    //
    unsigned radix;
    switch (istr.flags() & iob::basefield)
    {
        default:
            radix = 10U;
            if (!cf.is(ctype_facet::digit, c))
            {
                istr.setstate(iob::failbit);
                return istr;
            }
            break;

        case iob::oct:
            radix = 8U;
            if (!cf.is(ctype_facet::digit, c) || cf.narrow(c, '\0') > '7')
            {
                istr.setstate(iob::failbit);
                return istr;
            }
            break;

        case iob::hex:
            radix = 16U;
            if (!cf.is(ctype_facet::xdigit, c))
            {
                istr.setstate(iob::failbit);
                return istr;
            }
            if (cf.narrow(c, '\0') == '0')  // probably followed by 'x'
            {
                if (!istr.get(c))
                {
                    if (!istr.bad())
                    {
                        value.clear(); // we successfully read a '0' already
                        istr.clear();  // so the stream is still good()
                    }
                    return istr;
                }
                if (cf.narrow(cf.tolower(c), '\0') == 'x')
                {
                    if (!istr.get(c) || !cf.is(ctype_facet::xdigit, c))
                    {
                        istr.setstate(iob::failbit);
                        return istr;
                    }
                }
                else if (!cf.is(ctype_facet::xdigit, c))
                {
                    istr.putback(c); // non-digit follows valid input of '0'
                    if (istr.good())
                        value.clear();
                    return istr;
                }
            }
            break;
    }
    const integer big_radix(radix); // for multiplying

    using punct_facet = std::numpunct<Ch>;
    const punct_facet& pf = std::use_facet<punct_facet>(loc);

    const bool dont_allow_grouping =
        pf.grouping().empty() || pf.grouping()[0] == CHAR_MAX;
    const Ch sep = pf.thousands_sep();

    integer newval;

    do {
        if (cf.is(ctype_facet::digit, c))  // decimal digit, all radices
        {
            const char cc = cf.narrow(c, '\0');
            if (radix == 8U && cc > '7')  // non-digit successfully ends parse
                break;
            newval *= big_radix;
            newval += unsigned(cc - '0');
        }
        else if (radix == 16U && cf.is(ctype_facet::xdigit, c))
        {
            newval *= big_radix;
            newval += unsigned(cf.narrow(cf.tolower(c), '\0') - ('a' - 10));
        }
        else if (dont_allow_grouping || c != sep)
        {
            break;  // non-digit successfully ends parse
        }
        // Else just ignore thousands separators.
        // (We don't yet insist that they be in the right places.)
    } while (istr.get(c));

    if (istr.bad() || istr.fail() && !istr.eof())
        return istr;

    if (!istr.eof()) // Parse terminated by non-digit before EOF (normal).
    {
        istr.putback(c);
    }
    else // EOF is also OK.
    {
        istr.clear();
    }

    if (isneg)
    {
        newval.negate();
    }

    // Only after we're sure that everything worked:
    value = std::move(newval);

    return istr;
}

} // namespace bignum

#endif // INTEGER_IO_HPP_INCLUDED
